/*

  mux.c
  Mikael "MC" Cardell <mc@sics.se>
  1999-10-13
  
  This program multiplexes a number of packet streams and sends the
  packets out on a network interface or dumps it at a new tcpdump(8)
  savefile, or both.

  In particular, the program reads one or several tcpdump(8) savefiles
  and multiplexes the packets with regard to their timestamps,
  normalized to the file epoch (the timestamp of the first packet of
  each file). It is assumed that the timestamps of the tcpdump
  savefiles are allready in increasing order. If they are in the wrong
  order, the program might break.

  The program can perform IP number rewriting if given tuples of IP
  numbers as arguments.

  Run the program with options like this:
  
    -i interface

    -r(ewrite IP number tuples like this) a.b.c.d:e.f.g.h

    -f(ile) name

    inputfilename inputfilename ...

Copyright (c) 1999 Swedish Institute of Computer Science.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that: (1) source code
distributions retain the above copyright notice and this paragraph in
its entirety and (2) distributions including binary code include the
above copyright notice and this paragraph in its entirety in the
documentation or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

*/


/* Uncomment for very, very verbose debug messages */

#define DEBUG 1
int debug = 4;

#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/time.h>
#include <pcap.h>
#include <libnet.h>

#define print_tv(x) fprintf(stderr, " %ld.%.6ld\n", x tv_sec, x tv_usec);

/* A useful macro ripped from Matt Undy */

#define TIME_LT(x,y) (x tv_sec<y tv_sec||(x tv_sec==y tv_sec&&x tv_usec<y tv_usec))


/* Globals */

char *pname;
static char *version = "mux 19991005";
long failed_writes = 0,
    missed_intervals = 0,
    pkts_sent = 0,
    bytes_sent = 0;
struct timeval first_send = {-1, -1};
int tuplesno = 0;		/* total number of ip tuples for rewrite */

/* an array to store data and a packet from each flow */

struct flow
{
    struct pcap *infile;
    struct timeval epoch; /* epoch timestamp */
    struct pcap_pkthdr hdr;
    u_char *data;
    int eof; /* eof reached on the file */
} *flow;

/* array for IP tuples for rewriting */

struct iptuples
{
    struct in_addr
	from_ip,
	to_ip;
} *iptuples = NULL;

/* 
 *  Ethernet packet header prototype.  Too many OS's define this differently.
 */
struct etherhdr
{
#ifndef ETHER_ADDR_LEN
#define ETHER_ADDR_LEN 6
#endif
    u_char  ether_dhost[ETHER_ADDR_LEN];    /* destination ethernet address */
    u_char  ether_shost[ETHER_ADDR_LEN];    /* source ethernet address */
    u_short ether_type;                     /* packet type ID */
};


/* Functions */

/*
  Useful function ripped from Matt Undy.
  if op1>op2, return op1-op2 and 0
  if op1<op2, return |op1-op2| and 1
              (return a negative # with the sign bit in the return int)
  if op2>op1, subtract op2 from op1 and set return int
*/

int subtime(struct timeval *op1, struct timeval *op2, struct timeval *result)
{
    int borrow=0;     
    int sign=0;
    struct timeval *temp_time;
  
    if (TIME_LT(op1->, op2->))
    {
	sign = 1;
	temp_time = op1;
	op1 = op2;
	op2 = temp_time;
    }
    
    if (op1->tv_usec >= op2->tv_usec)
    {
	result->tv_usec = op1->tv_usec-op2->tv_usec;
    }
    else
    {
	result->tv_usec = (op1->tv_usec+1000000)-op2->tv_usec;
	borrow = 1;
    }
    result->tv_sec = (op1->tv_sec-op2->tv_sec)-borrow;
  
    return sign;
}

/*
  Rewrite IP-numbers and store away the packet
  */
void storepkt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
    struct timeval diff_ts;
    int i;
    u_int caplen = h->caplen; /* length of captured data */
    u_int len = h->len; /* recorded length of packet */
    struct etherhdr *ep;
    const u_char *datap;
    u_short ether_type;
    
    /* rewrite IP numbers */

    /* Is the packet captured smaller than an Ethernet header? */
    
    if (caplen < sizeof (struct etherhdr))
    {
	fprintf(stderr, "incomplete ethernet header. skipping...\n");
	return;
    }

    /* Move pass the Ether header, use datap as a pointer to the data
       as we move along. Keep sp pristine for later copying. */

    datap = sp;
    len -= sizeof(struct etherhdr);
    caplen -= sizeof(struct etherhdr);
    ep = (struct etherhdr *)datap;
    ether_type = ntohs(ep->ether_type);
    datap += sizeof(struct etherhdr);
    
    /* FIXME: We pretend 802.3 encapsulation doesn't exist for now. */

    if (ether_type == ETHERTYPE_IP)
    {
	const struct ip *ip;
	u_int hlen; /* IP header length */
	u_int ip_len;
	
	/* Ah, an IP packet! Just what we need. */

	ip = (const struct ip *)datap;
	
	/* Check for complete IP packet */

	if (len < sizeof (struct ip))
	{
	    fprintf(stderr, "truncated ip packet, length %d. skipping...\n",
		    len);
	    return;
	}

	hlen = ip->ip_hl * 4;
	ip_len = ntohs(ip->ip_len);
	if (len < ip_len)
	{
	    fprintf(stderr, "truncated-ip - %d bytes missing!\n",
		    ip_len - len);
	}
#ifdef DEBUG
	if (debug > 3)
	{
	    fprintf(stderr, "IP packet\n");
			
	    /* source and destination address */

	    fprintf(stderr, "\torig IP src: %s\n", inet_ntoa(ip->ip_src));
	    fprintf(stderr, "\torig IP dst: %s\n", inet_ntoa(ip->ip_dst));
	}
#endif

	/* Look for the IP numbers in the iptuples array. Rewrite if
           we find them. */

	for (i = 0; i < tuplesno; i ++)
	{
	    if (0 == memcmp(&ip->ip_src, &iptuples[i].from_ip,
		sizeof (struct in_addr)))
	    {
		memcpy(&ip->ip_src, &iptuples[i].to_ip,
		       sizeof (struct in_addr));
#ifdef DEBUG
		if (debug > 3)
		{
		    fprintf(stderr, "Rewriting IP number to %s\n",
			    inet_ntoa(iptuples[i].to_ip));
		}
#endif
	    }

	    if (0 == memcmp(&ip->ip_dst, &iptuples[i].from_ip,
		sizeof (struct in_addr)))
	    {
		memcpy(&ip->ip_dst, &iptuples[i].to_ip,
		       sizeof (struct in_addr));
#ifdef DEBUG
		if (debug > 3)
		{
		    fprintf(stderr, "Rewriting IP number to %s\n",
			    inet_ntoa(iptuples[i].to_ip));
		}
#endif
	    }
	}

#ifdef DEBUG
	if (debug > 3)
	{
	    fprintf(stderr, "\tIP src:      %s\n", inet_ntoa(ip->ip_src));
	    fprintf(stderr, "\tIP dst:      %s\n", inet_ntoa(ip->ip_dst));

	    fprintf(stderr, "\tTOS:         %d\n", ip->ip_tos);
	    fprintf(stderr, "\tLength:      %d\n", ntohs(ip->ip_len));
	    fprintf(stderr, "\tID:          %d\n", ntohs(ip->ip_id));
	    fprintf(stderr, "\tOffset:      %d\n", ntohs(ip->ip_off));
	    
	    fprintf(stderr, "\tTTL:         %d\n", ip->ip_ttl);
	    fprintf(stderr, "\tProtocol:    %d\n", ip->ip_p);
	    fprintf(stderr, "\tChecksum:    %d\n", ip->ip_sum);
	}
#endif	
	ip_len -= hlen;
    }
    else
    {
#if DEBUG
	if (debug > 2)
	{
	    fprintf(stderr, "unknown packet\n");
	}
#endif
    }
    
    /* Save the entire packet header in the packet array */
    
    memcpy(&flow[*user].hdr, h, sizeof (struct pcap_pkthdr));

    /* If file epoch is not set, this is the first packet of the
       file. Set the epoch. */
    
    if (flow[*user].epoch.tv_sec == -1
	&& flow[*user].epoch.tv_usec == -1)
    {
	flow[*user].epoch.tv_sec = h->ts.tv_sec;
	flow[*user].epoch.tv_usec = h->ts.tv_usec;

#ifdef DEBUG
	fprintf(stderr, "setting file %d epoch: %ld.%.6ld\n", *user,
		h->ts.tv_sec,
		h->ts.tv_usec);
#endif
    }

    /* Get the offset from the epoch */
    
    subtime(&h->ts, &flow[*user].epoch, &diff_ts);

#ifdef DEBUG
	fprintf(stderr, "packet from file %d, timestamp: %ld.%.6ld\n",
	       *user,
	       flow[*user].hdr.ts.tv_sec,
	       flow[*user].hdr.ts.tv_usec);

	fprintf(stderr, "\toffset %ld.%.6ld\n", diff_ts.tv_sec,
		diff_ts.tv_usec);
#endif
    
    /* Rewrite the timestamp in the packet header of the array */
    
    flow[*user].hdr.ts.tv_sec = diff_ts.tv_sec;
    flow[*user].hdr.ts.tv_usec = diff_ts.tv_usec;

    /* Get space for packet data and store it in the array */
    
    if ((flow[*user].data = (u_char *) malloc(h->caplen)) == NULL)
    { 
	fprintf(stderr, "%s: malloc: out of memory\n", pname);
	exit(EX_UNAVAILABLE);
    }

    memcpy(flow[*user].data, sp, h->caplen);
}

/* send a packet on the network interface after suitable pause, if
   necessary to keep intervals */

void writepkt(struct flow flow, struct link_int *write_if, char *interface)
{
    static struct timeval last_ts = {0, 0};
    struct timeval
	walltime,
	offset,
	pktinterval,
	wallinterval;
    int write_status = -1;
    register int i;
    
    /* Get the interval between this packet and last packet sent */

#ifdef DEBUG
    fprintf(stderr, "last_ts: ");
    print_tv(last_ts.);
#endif

    if (subtime(&flow.hdr.ts, &last_ts, &pktinterval) < 0)
    {
	/* negative time. packets are out of order. */
    }

#ifdef DEBUG
    fprintf(stderr, "\tpacket interval");
    print_tv(pktinterval.);
#endif

    /* store this packet timestamp as last timestamp */
    
    last_ts.tv_sec = flow.hdr.ts.tv_sec;
    last_ts.tv_usec = flow.hdr.ts.tv_usec;

#ifdef DEBUG
    fprintf(stderr, "setting new last_ts: ");
    print_tv(last_ts.);
#endif
    
    /* Get the current interval between the last sent packet and wall
       time */

    /* First, figure out what the offset is from the start of sending,
       or we can't compare timestamps */
       
    gettimeofday(&walltime, NULL);
    if (first_send.tv_sec == -1
	&& first_send.tv_usec == -1)
    {
	/* this is the first packet to be sent, our wall time epoch */

	first_send.tv_sec = walltime.tv_sec;
	first_send.tv_usec = walltime.tv_usec;
    }

    /* wait until wallinterval is equal or larger than pktinterval */

    /* NOTE: I used sleep() and usleep() here before. However, the
       implementation of usleep() is depending on the HZ constant in
       the BSD kernel, which is usually set to 100 Hz (10 ms). This is
       not enough for our purposes. Of course, this means this process
       will eat all the CPU available. */
       
    i = 0;
    do
    {
	gettimeofday(&walltime, NULL);
	if (subtime(&walltime, &first_send, &offset) < 0)
	{
	    /* the wall time is going backwards! someone may have set the
	       time on this machine under our feet. */
	}
	if (subtime(&offset, &last_ts, &wallinterval) < 0)
	{
	    /* the wall time is going backwards! someone may have set the
	       time on this machine under our feet. */
	}
#ifdef DEBUG
	i ++;
#endif
    } while (!timercmp(&wallinterval, &pktinterval, ==)
	     && !timercmp(&wallinterval, &pktinterval, >));

#ifdef DEBUG
    fprintf(stderr, "Looped %d times through busy-wait loop\n", i);
#endif
    
    while (write_status < 0)
    {
	write_status = libnet_write_link_layer(write_if, interface,
					flow.data, flow.hdr.caplen);
#ifdef DEBUG
	if (debug > 1) {
	    fprintf(stderr, "\twrite_status = %i\n", write_status);
	}
#endif
	if (write_status < 0) {
	    /*if errno =  = 55(ENOBUFS) loop, otherwise break*/
	    
	    if (errno != ENOBUFS)
	    {
		fprintf(stderr, "errno = %i\n", errno);
		perror("pcap_inject");
		exit(1);
	    }
	    else
	    {
		failed_writes ++;
	    }
	}
    } /* while */
    bytes_sent += write_status;
}

void usage(void)
{
    fprintf(stderr, "%s: Usage: %s [-i interface] [-f outfile] "
	    "[-r ip#1:ip#2] sourcefile1 [sourcefile2] "
	    "[...]\n", pname, pname);
    exit(EX_USAGE);
}

int main(int argc, char *argv[])
{
    pcap_dumper_t *dumpfile;
    char
	ch,			/* option index */
	ebuf[256],		/* general pcap error buffer */
	*outfilename = NULL,	/* output file, if any */
	*filename = NULL,	/* input filename */
	*interface = NULL,	/* output interface name */
	tostr[16],		/* new IP in dotted decimal  */
	fromstr[16];		/* original IP... */
    int rc,
	eofcount,		/* count of file ends */
	flowno;			/* total number of files/flows */

    long i;
    u_char
	files,			/* file counter */
	lowest;			/* pointer to file with lowest timestamp */
    struct link_int *write_if;	/* network interface */
    struct timeval
	now,			
	diff_ts;		/* difference in time */

    pname = argv[0];

    while ((ch = getopt(argc, argv, "i:f:r:v")) != -1)
    {
	switch (ch)
	{
	case 'i':
	    interface = optarg;
	    printf("interface == %s\n", interface);
	    break;
	case 'f':
	    outfilename = optarg;
	    break;
	case 'r':
	    /* Read a tuple of IP numbers in dotted decimal */
	    strcpy(fromstr, "not");
	    strcpy(tostr, "not");
	    sscanf(optarg, "%15[0-9.]:%15[0-9.]", fromstr, tostr);

	    /* Sanity check */

	    if (0 == strcmp(fromstr, "not") || 0 == strcmp(tostr, "not"))
	    {
		printf("%s: Not a complete IP tuple in rewriting rule.\n",
		       pname);
		usage();
		/* NOTREACHED */
	    }
	    /* Allocate space for the tuples array */
	    
	    if (NULL == (iptuples = (struct iptuples *)
		     realloc(iptuples, sizeof (struct iptuples)
			     * tuplesno + 1)))
	    {
		fprintf(stderr, "%s: realloc: out of memory\n",
			pname);
		exit(EX_UNAVAILABLE);
	    }

	    inet_aton(fromstr, &iptuples[tuplesno].from_ip);
	    inet_aton(tostr, &iptuples[tuplesno].to_ip);

	    /* Count this tuple too */
	    tuplesno ++;
	    
	    break;
	case 'v':
	    printf("Version: %s\n", version);
	    printf("Copyright (c) 1999 Swedish Institute of Computer Science.\n");
	    exit(0);
	    break;
	default:
	    printf("%s: Unknown argument", pname);
	    usage();
	    /* NOTREACHED */
	}
    }

    for (i = optind; i < argc; i ++)
    {
	filename = argv[i];
    }

    /* control arguments for sanity */

    if (NULL == interface && NULL == outfilename)
    {
	fprintf(stderr, "%s: no output specifier\n", pname);
	usage();
	/* NOTREACHED */
    }
    if (NULL == filename)
    {
	fprintf(stderr, "%s: no input file\n", pname);
	usage();
	/* NOTREACHED */
    }

#if 0
    /* FIXME: handle default interface */
    if (interface == NULL)
    {
	if (!(interface = pcap_lookupdev(ebuf)))
	{
	    fprintf(stderr, "can't determine default interface: %s\n", ebuf);
	    exit(EX_UNAVAILABLE);
	}
    }
#endif

#ifdef DEBUG
    fprintf(stderr, "interface == %s\n", interface);
#endif

    flowno = argc - optind;
    
    /* Get space for the per flow array */

    if ((flow = (struct flow *)
	 malloc((flowno) * sizeof (struct flow))) == NULL)
    { 
	fprintf(stderr, "%s: malloc: out of memory\n", pname);
	exit(EX_UNAVAILABLE);
    }

    /* Initialize the files */

    for (files = 0; files < flowno; files ++)
    {
	filename = argv[files + optind];

	/* Reset the epoch of each file */

	flow[files].epoch.tv_sec = -1;
	flow[files].epoch.tv_usec = -1;
	flow[files].eof = 0;
	
	flow[files].infile = pcap_open_offline(filename, ebuf);
	
	if (!flow[files].infile)
	{
	    fprintf(stderr, "%s: pcap_open_offline: %s\n", pname, ebuf);
	    exit(EX_NOINPUT);
	}

	if (pcap_dispatch(flow[files].infile, 1, (void *)&storepkt, &files)
	    == -1)
	{
	    pcap_perror(flow[files].infile, filename);
	}
    }
    
    /* Open the network interface */

    if (NULL != interface)
    {
	write_if = libnet_open_link_interface(interface, ebuf);
	if (write_if <= 0)
	{
	    fprintf(stderr, "Output interface: %s\n", ebuf);
	    exit(EX_OSERR);
	}
    }
        
    /* Open a new tcpdump(8) savefile */

    if (NULL != outfilename)
    {
	if ((dumpfile = pcap_dump_open(flow[0].infile, outfilename)) == NULL)
	{
	    printf("%s: pcap_dump_open: %s\n", pname,
		   pcap_geterr(flow[flowno].infile));
	    exit(EX_NOINPUT);
	}
    }

    /* loop until end of file in all source files */
    
    for (eofcount = 0; eofcount < flowno; )
    {
	/* compare all current timestamps with each other for the
           lowest time. */
	
	for (lowest = 0, i = 0; i < flowno; i ++)
	{
	    if (flow[i].eof != 1)
	    {
		if (timercmp(&flow[i].hdr.ts, &flow[lowest].hdr.ts, <)
		    || timercmp(&flow[i].hdr.ts, &flow[lowest].hdr.ts, ==))
		{
		    lowest = i;
		}
	    }
	}
	   
	/* Send packet on the interface and/or dump it on the file */

	pkts_sent ++;
	
	if (NULL != interface)
	{
	    writepkt(flow[lowest], write_if, interface);
	}

	if (NULL != outfilename)
	{
	    /* write packet to the dumpfile */
	    pcap_dump((u_char *)dumpfile, &flow[lowest].hdr,
		      flow[lowest].data);
	}
	
	/* free the space we allocated for the previous packet data */

	free(flow[lowest].data);
	
	/* read a new packet into the spot used previously by the
           lowest timestamp */
	
	rc = pcap_dispatch(flow[lowest].infile, 1, (void *)&storepkt,
			   &lowest);
	if (0 == rc)
	{
	    /* EOF in one of the source files. */
	    
	    flow[lowest].eof = 1;
	    eofcount ++;
	}
	else if (rc == -1)
	{
	    pcap_perror(flow[lowest].infile, filename);
	}

#ifdef DEBUG
	else
	{
	    fprintf(stderr, "\tnew timestamp: %ld.%.6ld\n",
		    flow[lowest].hdr.ts.tv_sec,
		    flow[lowest].hdr.ts.tv_usec);
	}
#endif
	
    } /* while not eof */

    if (NULL != interface)
    {
	libnet_close_link_interface(write_if);
    }
    
    if (NULL != outfilename)
    {
	pcap_dump_close(dumpfile);
    }

    gettimeofday(&now, NULL);
    if (subtime(&now, &first_send, &diff_ts) < 0)
    {
	/* negative time. */
    }

    if (NULL != interface)
    {
	printf("Finished transmission in %ld.%.6ld seconds.\n", diff_ts.tv_sec,
	       diff_ts.tv_usec);

	printf("Total number of bytes sent on the wire: %ld\n", bytes_sent);
	
	printf("(~%ld packets/second at %ld bytes/second, %ld bits/second)\n",
	       pkts_sent/diff_ts.tv_sec, bytes_sent/diff_ts.tv_sec,
	       8 * bytes_sent/diff_ts.tv_sec);
	printf("Failed writes: %ld\n", failed_writes);
	printf("Missed packet time intervals: %ld\n", missed_intervals);
    }

    printf("Total number of packets sent and/or written: %ld\n", pkts_sent);
    
    return 0;
}
