/*

  pktmux.c
  Mikael "MC" Cardell <mc@sics.se>
  1999-09-23
  
  This program multiplexes a number of packet streams together into
  one. It reads a number of tcpdump(8) savefiles and writes a new
  savefile for later replaying with a userland program such as
  tcpreplay(1) or a tight kernel loop.

  This is a version that keeps a pointer to where we are in a flow and
  holds the flow until the timestamps of the other files are all later
  or equal to the held timestamp. An earlier version read all
  timestamps first and sorted them.

Copyright (c) 1999 Swedish Institute of Computer Science.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that: (1) source code
distributions retain the above copyright notice and this paragraph in
its entirety and (2) distributions including binary code include the
above copyright notice and this paragraph in its entirety in the
documentation or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <pcap.h>

/* Uncomment for very, very verbose debug messages */

#define DEBUG 1

#ifdef SOLARIS

/* This useful macro is missing in Solaris sys/time.h */

#define	timersub(a, b, result)						      \
  do {									      \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;			      \
    if ((result)->tv_usec < 0) {					      \
      --(result)->tv_sec;						      \
      (result)->tv_usec += 1000000;					      \
    }									      \
  } while (0)

#endif

char *pname;

/* an array to store data and a packet from each flow */

struct flow
{
    struct pcap *infile;
    struct timeval epoch; /* epoch timestamp */
    struct pcap_pkthdr hdr;
    u_char *data;
    int eof; /* eof reached on the file */
} *flow;

void storepkt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
    struct timeval diff_ts;
    
    /* Save the entire packet header in the packet array */
    
    memcpy(&flow[*user].hdr, h, sizeof (struct pcap_pkthdr));

    /* If file epoch is not set, this is the first packet of the
       file. Set the epoch. */
    
    if (flow[*user].epoch.tv_sec == -1
	&& flow[*user].epoch.tv_usec == -1)
    {
	flow[*user].epoch.tv_sec = h->ts.tv_sec;
	flow[*user].epoch.tv_usec = h->ts.tv_usec;

#ifdef DEBUG
	printf("Epoch of file %d is %ld.%.6ld\n", *user, h->ts.tv_sec,
	       h->ts.tv_usec);
#endif
    }

    /* Get the offset from the epoch */
    
    timersub(&h->ts, &flow[*user].epoch, &diff_ts);

#ifdef DEBUG
    printf("Offset of this packet is %ld.%.6ld\n", diff_ts.tv_sec,
	   diff_ts.tv_usec);
#endif
    
    /* Rewrite the timestamp in the packet header of the array */
    
    flow[*user].hdr.ts.tv_sec = diff_ts.tv_sec;
    flow[*user].hdr.ts.tv_usec = diff_ts.tv_usec;

    /* Get space for packet data and store it in the array */
    
    if ((flow[*user].data = (u_char *) malloc(h->caplen)) == NULL)
    { 
	fprintf(stderr, "%s: malloc: out of memory\n", pname);
	exit(EX_UNAVAILABLE);
    }

    memcpy(flow[*user].data, sp, h->caplen);
}

int main(int argc, char *argv[])
{
    pcap_dumper_t *dumpfile;
    char ebuf[256];
    int rc, eofcount, flowno;
    long i;
    u_char files, lowest;
    char *filename, *dumpfilename;

    pname = argv[0];

    if (argc < 3)
    {
	fprintf(stderr, "%s: Usage: %s dumpfile sourcefile1 [sourcefile2] "
		"[...]\n", pname, pname);
	
	exit(EX_USAGE);
    }

    dumpfilename = argv[1];

#ifdef DEBUG
    printf("dumpfilename == %s\n", dumpfilename);
#endif

    flowno = argc - 2;
    
    /* Get space for the per flow array */

    if ((flow = (struct flow *)
	 malloc((flowno) * sizeof (struct flow))) == NULL)
    { 
	fprintf(stderr, "%s: malloc: out of memory\n", pname);
	exit(EX_UNAVAILABLE);
    }

    /* Initialize the files */

    for (files = 0; files < flowno; files ++)
    {
	filename = argv[files + 2];

#ifdef DEBUG
	printf("Processing file %s\n", filename);
#endif
	
	/* Reset the epoch of each file */

	flow[files].epoch.tv_sec = -1;
	flow[files].epoch.tv_usec = -1;
	flow[files].eof = 0;
	
	flow[files].infile = pcap_open_offline(filename, ebuf);
	
	if (!flow[files].infile)
	{
	    fprintf(stderr, "%s: pcap_open_offline: %s\n", pname, ebuf);
	    exit(EX_NOINPUT);
	}

	if (pcap_dispatch(flow[files].infile, 1, (void *)&storepkt, &files) == -1)
	{
	    pcap_perror(flow[files].infile, filename);
	}
    }
    
    /* Open a new tcpdump(8) savefile */

    if ((dumpfile = pcap_dump_open(flow[0].infile, dumpfilename)) == NULL)
    {
	printf("%s: pcap_dump_open: %s\n", pname,
	       pcap_geterr(flow[flowno].infile));
	exit(EX_NOINPUT);
    }

    /* loop until end of file in all source files */
    
    for (eofcount = 0; eofcount < flowno; )
    {
	/* compare all current timestamps with each other for the
           lowest time. */
	
	for (lowest = 0, i = 0; i < flowno; i ++)
	{
	    if (flow[i].eof != 1)
	    {
		if (timercmp(&flow[i].hdr.ts, &flow[lowest].hdr.ts, <)
		    || timercmp(&flow[i].hdr.ts, &flow[lowest].hdr.ts, ==))
		{
		    lowest = i;
		}
	    }
	}

	/* dump the packet with the lowest timestamp */
	
	/* The pcap(3) manual doesn't say that the first arg to
	   pcap_dump() is a FILE *, but it says so in the sources. And
	   no, the function doesn't return anything on a possible
	   error. Foo.
	   */

#ifdef DEBUG
	printf("Dumping the packet from flow %d with timestamp: %ld.%.6ld\n",
	       lowest,
	       flow[lowest].hdr.ts.tv_sec,
	       flow[lowest].hdr.ts.tv_usec);
#endif

	pcap_dump((u_char *)dumpfile, &flow[lowest].hdr, flow[lowest].data);

	/* free the space we allocated for the previous packet data */

	free(flow[lowest].data);
	
	/* read a new packet into the spot used previously by the
           lowest timestamp */
	
	rc = pcap_dispatch(flow[lowest].infile, 1, (void *)&storepkt,
			   &lowest);
	if (0 == rc)
	{
	    /* EOF in one of the source files. */
	    
	    flow[lowest].eof = 1;
	    eofcount ++;
	}
	else if (rc == -1)
	{
	    pcap_perror(flow[lowest].infile, filename);
	}

#ifdef DEBUG
	printf("New timestamp of packet from flow %d: %ld.%.6ld\n",
	       lowest,
	       flow[lowest].hdr.ts.tv_sec,
	       flow[lowest].hdr.ts.tv_usec);
#endif
	
    } /* while not eof */

    pcap_dump_close(dumpfile);
    
    return 0;
}
