/*
  shodiff.c
  Mikael "MC" Cardell <mc@sics.se>
  1999-08-23
  
  Show intervals between packets and/or offsets from an epoch (first
  timestamp) in a tcpdump(8) savefile.

Copyright (c) 1999 Swedish Institute of Computer Science.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that: (1) source code
distributions retain the above copyright notice and this paragraph in
its entirety and (2) distributions including binary code include the
above copyright notice and this paragraph in its entirety in the
documentation or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <unistd.h>
#include <sysexits.h>
#include <sys/types.h>
#include <pcap.h>


#define INTERVAL 1 /*0b001*/
#define OFFSET   2 /*0b010*/
#define VERBOSE  4 /*0b100*/

/* This useful macro is missing in Solaris */

#ifdef SOLARIS

#define	timersub(a, b, result)						      \
  do {									      \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;			      \
    if ((result)->tv_usec < 0) {					      \
      --(result)->tv_sec;						      \
      (result)->tv_usec += 1000000;					      \
    }									      \
  } while (0)

#endif

char *pname;
int mode = 0;

void pkthandler(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
    static int i = 0;
    static struct timeval last_ts = {-1, -1}, epoch;
    struct timeval diff_ts;
    
    i ++;

    if (mode & VERBOSE)
    {
	printf("Packet number: %d\n", i);

	printf("Arrived %ld.%.6ld with %d bytes (we captured %d)\n",
	       h->ts.tv_sec, h->ts.tv_usec, h->len, h->caplen);
    }

    if (mode & INTERVAL)
    {
	if (-1 != last_ts.tv_sec)
	{
	    timersub(&h->ts, &last_ts, &diff_ts);

	    if (mode & VERBOSE)
	    {
		printf("\t %ld.%.6ld seconds later than last packet.\n",
		       diff_ts.tv_sec, diff_ts.tv_usec);
	    }
	    else
	    {
		printf("%ld.%.6ld\n", diff_ts.tv_sec, diff_ts.tv_usec);
	    }
	}
	last_ts.tv_sec = h->ts.tv_sec;
	last_ts.tv_usec = h->ts.tv_usec;
    }

    if (mode & OFFSET)
    {
	if (1 == i)
	{
	    epoch.tv_sec = h->ts.tv_sec;
	    epoch.tv_usec = h->ts.tv_usec;
	}
	else
	{
	    timersub(&h->ts, &epoch, &diff_ts);
	    if (mode & VERBOSE)
	    {
		printf("\t offset: %ld seconds, %ld usecs.\n",
		       diff_ts.tv_sec, diff_ts.tv_usec);
	    }
	    else
	    {
		printf("%ld.%.6ld\n", diff_ts.tv_sec, diff_ts.tv_usec);
	    }
	}
    }
}

void usage(void)
{
    fprintf(stderr, "%s: Usage: %s [-i] [-o] [-v] filename\n", pname, pname);
    fprintf(stderr, "%s: -i for interval, -o for offset and -v for more verbosity\n", pname);
    exit(EX_USAGE);
}

int main(int argc, char **argv)
{
    char ebuf[256], *fname, ch;
    pcap_t *pd;

    pname = argv[0];

    while ((ch = getopt(argc, argv, "oiv")) != -1)
    {
	switch (ch)
	{
	case 'o':
	    mode = mode | OFFSET;
	    break;
	case 'i':
	    mode = mode | INTERVAL;
	    break;
	case 'v':
	    mode = mode | VERBOSE;
	    break;
	default:
	    usage();
	}
    }

    if (0 == mode || argc - optind != 1)
    {
	usage();
    }
    
    fname = argv[optind];

    if ((pd = pcap_open_offline(fname, ebuf)) == NULL)
    {
	printf("%s: pcap_open_offline: %s\n", pname, ebuf);
	exit(EX_NOINPUT);
    }

    /* cnt is -1, loop forever over the pcap_t handle pd until input
       is finished or an error occurs. */
    
    if (pcap_loop(pd, -1, pkthandler, 0) < 0)
    {
	fprintf(stderr, "%s: pcap_loop: %s\n", pname, pcap_geterr(pd));
	exit(1);
    }
    
    pcap_close(pd);
    return 0;
}
